@@ -25,6 +25,7 @@ from __future__ import absolute_import
 
 import itertools
 import warnings
+import copy
 import types
 import cmath
 
@@ -84,23 +85,24 @@ class efield(object):
         """initialize field container"""
         self.field_generator = field_generator
         self.wavelengths = np.array(wavelengths)
-        self.kwargs = kwargs
+        self.kwargs = copy.deepcopy(kwargs)
+        _kwargs = copy.deepcopy(kwargs)
         
         ## --- generate parameter-sets for field-generator
-        if type(kwargs) == dict:
+        if type(_kwargs) == dict:
             ## --- integer parameters to list
-            for key in kwargs:
-                if type(kwargs[key]) not in [list, np.ndarray]:
-                    kwargs[key] = [kwargs[key]]
+            for key in _kwargs:
+                if type(_kwargs[key]) not in [list, np.ndarray]:
+                    _kwargs[key] = [_kwargs[key]]
             
             ## --- generate all permutations of kwargs for direct use in field_generator
-            varNames = sorted(kwargs)
+            varNames = sorted(_kwargs)
             self.kwargs_permutations = [dict(zip(varNames, prod)) for 
-                                    prod in itertools.product(*(kwargs[varName] 
+                                    prod in itertools.product(*(_kwargs[varName] 
                                                 for varName in varNames))]
-        elif type(kwargs) == list:
+        elif type(_kwargs) == list:
             self.kwargs_permutations = []
-            for kw in kwargs:
+            for kw in _kwargs:
                 self.kwargs_permutations.append(kw)
                 if type(kw) != dict:
                     raise ValueError("Wrong input for 'kwargs': Must be either dict or list of dicts.")
@@ -1099,7 +1101,7 @@ def fast_electron(pos, env_dict, wavelength,
     if 'eps_env' in env_dict.keys():
         n2 = n2 = n3 = env_dict['eps_env']**0.5
     else:
-        n1 = env_dict['eps1']**0.5
+        # n1 = env_dict['eps1']**0.5
         n2 = env_dict['eps2']**0.5
         n3 = env_dict['eps3']**0.5
         # spacing = np.float32(env_dict['spacing'].real)
@@ -1152,6 +1154,957 @@ def fast_electron(pos, env_dict, wavelength,
 
 
 
+#************************************************************************
+#************************************************************************
+# Definition of HG, radially or azimuthally pol. doughnut modes following
+#        Novotny & Hecht Principle of Nano-Optics - p62
+#
+# Implementation by A. Arbouet, CEMES-CNRS, 2021
+#
+#************************************************************************
+#************************************************************************               
+def _fw(theta, f0, thetamax):
+
+    """ Novotny & Hecht Principle of Nano-Optics - p62
+
+    Parameters
+    ----------
+    theta : angle
+    f0: lens filling factor (Novotny p62) f0 = w0/ (f sin theta max) 
+    thetamax: max. angle allowed in lens
+    Returns
+    -------     
+    
+    Notes
+    -----
+         add notes
+    
+    """
+    
+    fres = np.exp(-(np.sin(theta)/(f0*np.sin(thetamax)))**2)
+    return fres                  
+                    
+# conversion cartersian to cylindrical
+def _cart_cyl(x,y):
+    rho = np.sqrt(x**2 + y**2)
+    phi = np.arctan2(y,x)
+    return rho,phi   
+    
+def _I00(x,y,z,k,n,NA,f,w0):
+
+    """ Novotny & Hecht Principle of Nano-Optics - p62
+
+    Parameters
+    ----------
+    NA : numerical aperture
+    
+    Returns
+    -------     
+    
+    Notes
+    -----
+         add notes
+    
+    """
+    import scipy.integrate as integrate
+    import scipy.special as special
+    
+    thetamax = np.arcsin(NA/n)
+    f0 = w0/(f*np.sin(thetamax)) 
+    rho, phi = _cart_cyl(x,y)    
+    real = integrate.quad(lambda theta: np.real(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)*(1+np.cos(theta))*special.j0(k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    imag = integrate.quad(lambda theta: np.imag(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)*(1+np.cos(theta))*special.j0(k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    
+    return real[0] + 1j*imag[0] 
+          
+def _I01(x,y,z,k,n,NA,f,w0):
+
+    """ Novotny & Hecht Principle of Nano-Optics - p62
+
+    Parameters
+    ----------
+    NA : numerical aperture
+    
+    Returns
+    -------     
+    
+    Notes
+    -----
+         add notes
+    
+    """
+    import scipy.integrate as integrate
+    import scipy.special as special
+    
+    thetamax = np.arcsin(NA/n)
+    f0 = w0/(f*np.sin(thetamax))    
+    rho, phi = _cart_cyl(x,y)    
+    real = integrate.quad(lambda theta: np.real(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**2*special.j1(k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    imag = integrate.quad(lambda theta: np.imag(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**2*special.j1(k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    
+    return real[0] + 1j*imag[0] 
+
+def _I02(x,y,z,k,n,NA,f,w0):
+
+    """ Novotny & Hecht Principle of Nano-Optics - p62
+
+    Parameters
+    ----------
+    NA : numerical aperture
+    
+    Returns
+    -------     
+    
+    Notes
+    -----
+         add notes
+    
+    """
+    import scipy.integrate as integrate
+    import scipy.special as special
+    
+    thetamax = np.arcsin(NA/n)
+    f0 = w0/(f*np.sin(thetamax))
+    rho, phi = _cart_cyl(x,y)    
+    real = integrate.quad(lambda theta: np.real(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)*(1-np.cos(theta))*special.jv(2,k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    imag = integrate.quad(lambda theta: np.imag(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)*(1-np.cos(theta))*special.jv(2,k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    
+    return real[0] + 1j*imag[0] 
+          
+
+def _I10(x,y,z,k,n,NA,f,w0):
+
+    """ Novotny & Hecht Principle of Nano-Optics - p62
+
+    Parameters
+    ----------
+    NA : numerical aperture
+    
+    Returns
+    -------     
+    
+    Notes
+    -----
+         add notes
+    
+    """
+    import scipy.integrate as integrate
+    import scipy.special as special
+    
+    thetamax = np.arcsin(NA/n)
+    f0 = w0/(f*np.sin(thetamax))    
+    rho, phi = _cart_cyl(x,y)    
+    real = integrate.quad(lambda theta: np.real(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**3*special.j0(k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    imag = integrate.quad(lambda theta: np.imag(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**3*special.j0(k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    
+    return real[0] + 1j*imag[0] 
+
+
+def _I11(x,y,z,k,n,NA,f,w0):
+
+    """ Novotny & Hecht Principle of Nano-Optics - p62
+
+    Parameters
+    ----------
+    NA : numerical aperture
+    
+    Returns
+    -------     
+    
+    Notes
+    -----
+         add notes
+    
+    """
+    import scipy.integrate as integrate
+    import scipy.special as special
+    
+    thetamax = np.arcsin(NA/n)
+    f0 = w0/(f*np.sin(thetamax))     
+    rho, phi = _cart_cyl(x,y)    
+    real = integrate.quad(lambda theta: np.real(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**2*(1+3*np.cos(theta))*special.j1(k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    imag = integrate.quad(lambda theta: np.imag(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**2*(1+3*np.cos(theta))*special.j1(k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    
+    return real[0] + 1j*imag[0] 
+
+
+def _I12(x,y,z,k,n,NA,f,w0):
+
+    """ Novotny & Hecht Principle of Nano-Optics - p62
+
+    Parameters
+    ----------
+    NA : numerical aperture
+    
+    Returns
+    -------     
+    
+    Notes
+    -----
+         add notes
+    
+    """
+    import scipy.integrate as integrate
+    import scipy.special as special
+    
+    thetamax = np.arcsin(NA/n)
+    f0 = w0/(f*np.sin(thetamax))     
+    rho, phi = _cart_cyl(x,y)    
+    real = integrate.quad(lambda theta: np.real(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**2*(1-np.cos(theta))*special.j1(k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    imag = integrate.quad(lambda theta: np.imag(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**2*(1-np.cos(theta))*special.j1(k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    
+    return real[0] + 1j*imag[0] 
+
+def _I13(x,y,z,k,n,NA,f,w0):
+
+    """ Novotny & Hecht Principle of Nano-Optics - p62
+
+    Parameters
+    ----------
+    NA : numerical aperture
+    
+    Returns
+    -------     
+    
+    Notes
+    -----
+         add notes
+    
+    """
+    import scipy.integrate as integrate
+    import scipy.special as special
+    
+    thetamax = np.arcsin(NA/n)
+    f0 = w0/(f*np.sin(thetamax)) 
+    rho, phi = _cart_cyl(x,y)    
+    real = integrate.quad(lambda theta: np.real(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**3*special.jv(2,k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    imag = integrate.quad(lambda theta: np.imag(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**3*special.jv(2,k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    
+    return real[0] + 1j*imag[0] 
+
+def _I14(x,y,z,k,n,NA,f,w0):
+
+    """ Novotny & Hecht Principle of Nano-Optics - p62
+
+    Parameters
+    ----------
+    NA : numerical aperture
+    
+    Returns
+    -------     
+    
+    Notes
+    -----
+         add notes
+    
+    """
+    import scipy.integrate as integrate
+    import scipy.special as special
+    
+    thetamax = np.arcsin(NA/n)
+    f0 = w0/(f*np.sin(thetamax))    
+    rho, phi = _cart_cyl(x,y)    
+    real = integrate.quad(lambda theta: np.real(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**2*(1-np.cos(theta))*special.jv(3,k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    imag = integrate.quad(lambda theta: np.imag(_fw(theta,f0,thetamax)*np.sqrt(np.cos(theta))
+                         *np.sin(theta)**2*(1-np.cos(theta))*special.jv(3,k*rho*np.sin(theta))
+                         *np.exp(1j*k*z*np.cos(theta))), 0, thetamax)
+    
+    return real[0] + 1j*imag[0] 
+
+
+def HermiteGauss00(pos, env_dict, wavelength, #theta=None, polarization_state=None, 
+             xSpot=0.0, ySpot=0.0, zSpot=0.0, kSign=-1.0,
+             NA=0.5, f=100, w0=1, returnField='E'):
+
+    """Gaussian TEM mode (0,0)
+    
+    Focal fields - Expressions from Novotny & Hecht - Principle of Nano-Optics (3.6)
+    Generalized to take into account the presence of an interface
+    
+    Authors: A. Arbouet, Y. Brûlé, G. Colas-des-Francs, 2021
+    
+    
+    Parameters
+    ----------
+    pos : np.array
+            list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
+    
+    env_dict : dict
+                    Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
+                    description of environment. Must contain either "eps_env" or ["eps2"].
+    
+    wavelength : float
+        Wavelength in nm
+
+   theta : float, default: None
+        either 'theta' or 'polarization_state' must be given.
+        linear polarization angle in degrees, 0deg = 0X direction
+    
+    polarization_state : 4-tuple of float, default: None
+        either 'theta' or 'polarization_state' must be given.
+        polarization state with field amplitudes and phases, tuple of 4 float:
+        (E0x, E0y, Dphi, Aphi): E0X amplitde, E0Y amplitde, phase difference 
+        between X and Y components (in rad), absolute phase of plane wave (in rad).
+        The field is then calculated as E = (E0x, E0y*exp(i*Dphi*z), 0)*exp(i*Aphi*z).
+            Dphi : 
+                - positive: left hand rotating polarization
+                - negative: right hand rotating polarization 
+                - example: left circular pol. with (E0x=1, E0y=1, Dphi=np.pi/2., phi=0)
+
+    xSpot, ySpot, zSpot : float, float, float, default: 0, 0, 0
+                                       focal spot position (in nm)
+
+    kSign : int, default: -1
+               sign of wavenumber. 
+               +1: propagation from bottom to top (towards increasing z)
+               -1: propagation from top to bottom (towards smaller z, default) 
+
+      NA : float, default: 0.5
+              lens numerical aperture (NA = n sin thetamax),
+
+      f :     float, default: 100
+              lens focal distance (mm)
+
+      w0 : float, default: 1 
+              beam waist  (mm)
+
+      returnField : str, default: 'E'
+              if 'E': returns electric field; if 'B': magnetic field
+    
+    Returns
+    -------
+      E0 (B0):       Complex E-(B-)Field at each dipole position as 
+                      list of (complex) 3-tuples: [(Ex1, Ey1, Ez1), ...]    
+
+    """
+    
+    
+    if 'eps_env' in env_dict.keys():
+        cn1 = cn2 = env_dict['eps_env']**0.5
+    else:
+        cn1 = env_dict['eps1']**0.5
+        cn2 = env_dict['eps2']**0.5
+        cn3 = env_dict['eps3']**0.5
+        # spacing = env_dict['spacing']**0.5
+#        if cn1 != cn2 or cn2 != cn3:
+    if np.imag(cn1) != 0 or np.imag(cn2) != 0 or np.imag(cn3) != 0:
+        warnings.warn("Special modes only support real refractive index")
+    if cn1 != cn2 or cn2 != cn3:
+        warnings.warn("Special modes don't take reflection / transmission at interfaces into account. Results might be inaccurate.")
+    
+    cn1 = np.real(cn1)
+    cn2 = np.real(cn2)
+    cn3 = np.real(cn3)
+        
+    f *= 1E6  # conversion mm > nm
+    w0 *= 1E6  # conversion mm > nm
+        
+    xm, ym, zm = np.transpose(pos)    
+    npts = len(xm)
+ 
+    if kSign == -1 : 
+        n = cn2
+    else:    
+        n = cn1
+    k = kSign * n * (2*np.pi / wavelength)
+    prefactor = 0.5*1j*k*f*np.sqrt(cn1/cn2)*np.exp(-1j*k*f)
+
+    Ex = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Ey = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Ez = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Bx = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    By = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Bz = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    
+    for ipt in range(npts):
+        
+        x = xm[ipt] - xSpot
+        y = ym[ipt] - ySpot
+        z = zm[ipt] - zSpot
+       
+        rho, phi = _cart_cyl(x, y)
+           
+        I_00 = _I00(x,y,z,k,n,NA,f,w0)
+        I_02 = _I02(x,y,z,k,n,NA,f,w0)
+        I_01 = _I01(x,y,z,k,n,NA,f,w0)
+     
+# Novotny (3.66)
+# Electric field
+        Ex[ipt] = prefactor * (I_00 + I_02*np.cos(2*phi))
+        Ey[ipt] = prefactor * I_02 * np.sin(2*phi)
+        Ez[ipt] = prefactor * (-2 * 1j * I_01 * np.cos(phi) )    
+
+# Magnetic field
+
+    if returnField == 'B':
+        output = np.asfortranarray(np.transpose([Bx, By, Bz]))
+    else :
+        output = np.asfortranarray(np.transpose([Ex, Ey, Ez]))
+    ## --- return as fortran array to avoid copying of arrays in memory
+    return output
+
+
+def HermiteGauss10(pos, env_dict, wavelength, #theta=None, polarization_state=None, 
+             xSpot=0.0, ySpot=0.0, zSpot=0.0, kSign=-1.0,
+             NA=0.5, f=100, w0=1, returnField='E'):
+
+    """Gaussian TEM mode (1,0)
+    
+    Focal fields - Expressions from Novotny & Hecht - Principle of Nano-Optics (3.6)
+    Generalized to take into account the presence of an interface
+    
+    Authors: A. Arbouet, Y. Brûlé, G. Colas-des-Francs, 2021
+    
+    
+    Parameters
+    ----------
+    pos : np.array
+            list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
+    
+    env_dict : dict
+                    Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
+                    description of environment. Must contain either "eps_env" or ["eps2"].
+    
+    wavelength : float
+        Wavelength in nm
+
+   theta : float, default: None
+        either 'theta' or 'polarization_state' must be given.
+        linear polarization angle in degrees, 0deg = 0X direction
+    
+    polarization_state : 4-tuple of float, default: None
+        either 'theta' or 'polarization_state' must be given.
+        polarization state with field amplitudes and phases, tuple of 4 float:
+        (E0x, E0y, Dphi, Aphi): E0X amplitde, E0Y amplitde, phase difference 
+        between X and Y components (in rad), absolute phase of plane wave (in rad).
+        The field is then calculated as E = (E0x, E0y*exp(i*Dphi*z), 0)*exp(i*Aphi*z).
+            Dphi : 
+                - positive: left hand rotating polarization
+                - negative: right hand rotating polarization 
+                - example: left circular pol. with (E0x=1, E0y=1, Dphi=np.pi/2., phi=0)
+
+    xSpot, ySpot, zSpot : float, float, float, default: 0, 0, 0
+                                       focal spot position (in nm)
+
+    kSign : int, default: -1
+               sign of wavenumber. 
+               +1: propagation from bottom to top (towards increasing z)
+               -1: propagation from top to bottom (towards smaller z, default) 
+
+      NA : float, default: 0.5
+              lens numerical aperture (NA = n sin thetamax),
+
+      f :     float, default: 100
+              lens focal distance (mm)
+
+      w0 : float, default: 1 
+              beam waist  (mm)
+
+      returnField : str, default: 'E'
+              if 'E': returns electric field; if 'B': magnetic field
+    
+    Returns
+    -------
+      E0 (B0):       Complex E-(B-)Field at each dipole position as 
+                      list of (complex) 3-tuples: [(Ex1, Ey1, Ez1), ...]    
+
+    """
+    
+    if 'eps_env' in env_dict.keys():
+        cn1 = cn2 = env_dict['eps_env']**0.5
+    else:
+        cn1 = env_dict['eps1']**0.5
+        cn2 = env_dict['eps2']**0.5
+        cn3 = env_dict['eps3']**0.5
+        # spacing = env_dict['spacing']**0.5
+#        if cn1 != cn2 or cn2 != cn3:
+    if np.imag(cn1) != 0 or np.imag(cn2) != 0 or np.imag(cn3) != 0:
+        warnings.warn("Special modes only support real refractive index")
+    if cn1 != cn2 or cn2 != cn3:
+        warnings.warn("Special modes don't take reflection / transmission at interfaces into account. Results might be inaccurate.")
+    
+    
+    cn1 = np.real(cn1)
+    cn2 = np.real(cn2)
+    cn3 = np.real(cn3)
+        
+    f *= 1E6  # conversion mm > nm
+    w0 *= 1E6  # conversion mm > nm
+        
+    xm, ym, zm = np.transpose(pos)    
+    npts = len(xm)
+
+    if kSign == -1 : 
+        n = cn2
+    else:    
+        n = cn1
+    k = kSign * n * (2*np.pi / wavelength)
+    prefactor = 0.5*1j*k*f**2*np.sqrt(cn1/cn2)*np.exp(-1j*k*f)/w0
+
+    Ex = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Ey = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Ez = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Bx = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    By = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Bz = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    
+    for ipt in range(npts):
+        
+        x = xm[ipt] - xSpot
+        y = ym[ipt] - ySpot
+        z = zm[ipt] - zSpot
+    
+        rho, phi = _cart_cyl(x, y)
+           
+        I_11 = _I11(x,y,z,k,n,NA,f,w0)
+        I_14 = _I14(x,y,z,k,n,NA,f,w0)
+        I_12 = _I12(x,y,z,k,n,NA,f,w0)
+        I_10 = _I10(x,y,z,k,n,NA,f,w0)
+        I_13 = _I13(x,y,z,k,n,NA,f,w0)
+     
+# Novotny (3.66)
+# Electric field
+        Ex[ipt] = prefactor * ( 1j*I_11*np.cos(phi) + 1j * I_14*np.cos(3*phi))
+        Ey[ipt] = prefactor * (-1j*I_12*np.sin(phi) + 1j * I_14*np.sin(3*phi))
+        Ez[ipt] = prefactor * (-2 * I_10 + 2 * I_13 * np.cos(2*phi) )    
+
+# Magnetic field
+
+    if returnField == 'B':
+        output = np.asfortranarray(np.transpose([Bx, By, Bz]))
+    else :
+        output = np.asfortranarray(np.transpose([Ex, Ey, Ez]))
+    ## --- return as fortran array to avoid copying of arrays in memory
+    return output
+
+
+def HermiteGauss01(pos, env_dict, wavelength, #theta=None, polarization_state=None, 
+             xSpot=0.0, ySpot=0.0, zSpot=0.0, kSign=-1.0,
+             NA=0.5, f=100, w0=1, returnField='E'):
+
+    """Gaussian TEM mode (0,1)
+    
+    Focal fields - Expressions from Novotny & Hecht - Principle of Nano-Optics (3.6)
+    Generalized to take into account the presence of an interface
+    
+    Authors: A. Arbouet, Y. Brûlé, G. Colas-des-Francs, 2021
+    
+    
+    Parameters
+    ----------
+    pos : np.array
+            list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
+    
+    env_dict : dict
+                    Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
+                    description of environment. Must contain either "eps_env" or ["eps2"].
+    
+    wavelength : float
+        Wavelength in nm
+
+   theta : float, default: None
+        either 'theta' or 'polarization_state' must be given.
+        linear polarization angle in degrees, 0deg = 0X direction
+    
+    polarization_state : 4-tuple of float, default: None
+        either 'theta' or 'polarization_state' must be given.
+        polarization state with field amplitudes and phases, tuple of 4 float:
+        (E0x, E0y, Dphi, Aphi): E0X amplitde, E0Y amplitde, phase difference 
+        between X and Y components (in rad), absolute phase of plane wave (in rad).
+        The field is then calculated as E = (E0x, E0y*exp(i*Dphi*z), 0)*exp(i*Aphi*z).
+            Dphi : 
+                - positive: left hand rotating polarization
+                - negative: right hand rotating polarization 
+                - example: left circular pol. with (E0x=1, E0y=1, Dphi=np.pi/2., phi=0)
+
+    xSpot, ySpot, zSpot : float, float, float, default: 0, 0, 0
+                                       focal spot position (in nm)
+
+    kSign : int, default: -1
+               sign of wavenumber. 
+               +1: propagation from bottom to top (towards increasing z)
+               -1: propagation from top to bottom (towards smaller z, default) 
+
+      NA : float, default: 0.5
+              lens numerical aperture (NA = n sin thetamax),
+
+      f :     float, default: 100
+              lens focal distance (mm)
+
+      w0 : float, default: 1 
+              beam waist  (mm)
+
+      returnField : str, default: 'E'
+              if 'E': returns electric field; if 'B': magnetic field
+    
+    Returns
+    -------
+      E0 (B0):       Complex E-(B-)Field at each dipole position as 
+                      list of (complex) 3-tuples: [(Ex1, Ey1, Ez1), ...]    
+
+    """
+    
+    
+    if 'eps_env' in env_dict.keys():
+        cn1 = cn2 = env_dict['eps_env']**0.5
+    else:
+        cn1 = env_dict['eps1']**0.5
+        cn2 = env_dict['eps2']**0.5
+        cn3 = env_dict['eps3']**0.5
+        # spacing = env_dict['spacing']**0.5
+#        if cn1 != cn2 or cn2 != cn3:
+    if np.imag(cn1) != 0 or np.imag(cn2) != 0 or np.imag(cn3) != 0:
+        warnings.warn("Special modes only support real refractive index")
+    if cn1 != cn2 or cn2 != cn3:
+        warnings.warn("Special modes don't take reflection / transmission at interfaces into account. Results might be inaccurate.")
+    
+    cn1 = np.real(cn1)
+    cn2 = np.real(cn2)
+    cn3 = np.real(cn3)
+        
+    f *= 1E6  # conversion mm > nm
+    w0 *= 1E6  # conversion mm > nm
+        
+    xm, ym, zm = np.transpose(pos)    
+    npts = len(xm)
+    if kSign == -1 : 
+        n = cn2
+    else:    
+        n = cn1
+    k = kSign * n * (2*np.pi / wavelength)
+    prefactor = 0.5*1j*k*f**2*np.sqrt(cn1/cn2)*np.exp(-1j*k*f)/w0
+
+    Ex = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Ey = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Ez = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Bx = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    By = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Bz = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    
+    for ipt in range(npts):
+        
+        x = xm[ipt] - xSpot
+        y = ym[ipt] - ySpot
+        z = zm[ipt] - zSpot
+    
+        rho, phi = _cart_cyl(x, y)
+           
+        I_11 = _I11(x,y,z,k,n,NA,f,w0)
+        I_14 = _I14(x,y,z,k,n,NA,f,w0)
+        I_12 = _I12(x,y,z,k,n,NA,f,w0)
+        I_13 = _I13(x,y,z,k,n,NA,f,w0)
+        
+# Novotny (3.66)
+# Electric field
+        Ex[ipt] = prefactor * ( 1j*(I_11 + 2*I_12)*np.sin(phi) + 1j * I_14*np.sin(3*phi))
+        Ey[ipt] = prefactor * (-1j*I_12*np.cos(phi) - 1j * I_14*np.cos(3*phi))
+        Ez[ipt] = prefactor * ( 2 * I_13 * np.sin(2*phi) )    
+
+# Magnetic field
+
+    if returnField == 'B':
+        output = np.asfortranarray(np.transpose([Bx, By, Bz]))
+    else :
+        output = np.asfortranarray(np.transpose([Ex, Ey, Ez]))
+    ## --- return as fortran array to avoid copying of arrays in memory
+    return output
+                 
+
+def Radial_pol_doughnut(pos, env_dict, wavelength, #theta=None, polarization_state=None, 
+                        xSpot=0.0, ySpot=0.0, zSpot=0.0, kSign=-1.0,
+                        NA=0.5, f=100, w0=1, returnField='E'):
+
+    """radially polarized doughnut mode
+
+    Focal fields - Expressions from Novotny & Hecht - Principle of Nano-Optics (3.6)
+    Generalized to take into account the presence of an interface
+    
+    Authors: A. Arbouet, Y. Brûlé, G. Colas-des-Francs, 2021
+    
+    
+    Parameters
+    ----------
+    pos : np.array
+            list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
+    
+    env_dict : dict
+                    Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
+                    description of environment. Must contain either "eps_env" or ["eps2"].
+    
+    wavelength : float
+        Wavelength in nm
+
+   theta : float, default: None
+        either 'theta' or 'polarization_state' must be given.
+        linear polarization angle in degrees, 0deg = 0X direction
+    
+    polarization_state : 4-tuple of float, default: None
+        either 'theta' or 'polarization_state' must be given.
+        polarization state with field amplitudes and phases, tuple of 4 float:
+        (E0x, E0y, Dphi, Aphi): E0X amplitde, E0Y amplitde, phase difference 
+        between X and Y components (in rad), absolute phase of plane wave (in rad).
+        The field is then calculated as E = (E0x, E0y*exp(i*Dphi*z), 0)*exp(i*Aphi*z).
+            Dphi : 
+                - positive: left hand rotating polarization
+                - negative: right hand rotating polarization 
+                - example: left circular pol. with (E0x=1, E0y=1, Dphi=np.pi/2., phi=0)
+
+    xSpot, ySpot, zSpot : float, float, float, default: 0, 0, 0
+                                       focal spot position (in nm)
+
+    kSign : int, default: -1
+               sign of wavenumber. 
+               +1: propagation from bottom to top (towards increasing z)
+               -1: propagation from top to bottom (towards smaller z, default) 
+
+      NA : float, default: 0.5
+              lens numerical aperture (NA = n sin thetamax),
+
+      f :     float, default: 100
+              lens focal distance (mm)
+
+      w0 : float, default: 1 
+              beam waist  (mm)
+
+      returnField : str, default: 'E'
+              if 'E': returns electric field; if 'B': magnetic field
+    
+    Returns
+    -------
+      E0 (B0):       Complex E-(B-)Field at each dipole position as 
+                      list of (complex) 3-tuples: [(Ex1, Ey1, Ez1), ...]    
+
+    """
+    
+    
+    if 'eps_env' in env_dict.keys():
+        cn1 = cn2 = env_dict['eps_env']**0.5
+    else:
+        cn1 = env_dict['eps1']**0.5
+        cn2 = env_dict['eps2']**0.5
+        cn3 = env_dict['eps3']**0.5
+        # spacing = env_dict['spacing']**0.5
+#        if cn1 != cn2 or cn2 != cn3:
+    if np.imag(cn1) != 0 or np.imag(cn2) != 0 or np.imag(cn3) != 0:
+        warnings.warn("Special modes only support real refractive index")
+    if cn1 != cn2 or cn2 != cn3:
+        warnings.warn("Special modes don't take reflection / transmission at interfaces into account. Results might be inaccurate.")
+    
+    cn1 = np.real(cn1)
+    cn2 = np.real(cn2)
+    cn3 = np.real(cn3)
+        
+    f *= 1E6  # conversion mm > nm
+    w0 *= 1E6  # conversion mm > nm
+        
+    xm, ym, zm = np.transpose(pos)    
+    npts = len(xm)
+    if kSign == -1 : 
+        n = cn2
+    else:    
+        n = cn1
+    k = kSign * n * (2*np.pi / wavelength)
+    prefactor = 0.5*1j*k*f**2*np.sqrt(cn1/cn2)*np.exp(-1j*k*f)/w0
+
+    Ex = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Ey = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Ez = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Bx = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    By = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Bz = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    
+    for ipt in range(npts):
+        
+        x = xm[ipt] - xSpot
+        y = ym[ipt] - ySpot
+        z = zm[ipt] - zSpot
+    
+        rho, phi = _cart_cyl(x, y)
+           
+        I_10 = _I10(x,y,z,k,n,NA,f,w0)
+        I_11 = _I11(x,y,z,k,n,NA,f,w0)
+        I_12 = _I12(x,y,z,k,n,NA,f,w0)
+     
+# Novotny (3.66)
+# Electric field
+        Ex[ipt] = prefactor * ( 1j*(I_11 - I_12)*np.cos(phi) )
+        Ey[ipt] = prefactor * ( 1j*(I_11 - I_12)*np.sin(phi))
+        Ez[ipt] = prefactor * ( -4 * I_10)    
+
+# Magnetic field
+
+    if returnField == 'B':
+        output = np.asfortranarray(np.transpose([Bx, By, Bz]))
+    else :
+        output = np.asfortranarray(np.transpose([Ex, Ey, Ez]))
+    ## --- return as fortran array to avoid copying of arrays in memory
+    return output    
+
+
+def Azimuth_pol_doughnut(pos, env_dict, wavelength, #theta=None, polarization_state=None, 
+             xSpot=0.0, ySpot=0.0, zSpot=0.0, kSign=-1.0,
+             NA=0.5, f=100, w0=1, returnField='E'):
+    """Azimuthally polarized doughnut mode
+
+    Focal fields - Expressions from Novotny & Hecht - Principle of Nano-Optics (3.6)
+    Generalized to take into account the presence of an interface
+    
+    Authors: A. Arbouet, Y. Brûlé, G. Colas-des-Francs, 2021
+    
+    
+    Parameters
+    ----------
+    pos : np.array
+            list of 3-tuple coordinates to evaluate field at: [[x1,y1,z1], [x2,y2,z2], ... ]
+    
+    env_dict : dict
+                    Must be compatible with `sim.dyads.getConfigDictG` typed numba dict.
+                    description of environment. Must contain either "eps_env" or ["eps2"].
+    
+    wavelength : float
+        Wavelength in nm
+
+   theta : float, default: None
+        either 'theta' or 'polarization_state' must be given.
+        linear polarization angle in degrees, 0deg = 0X direction
+    
+    polarization_state : 4-tuple of float, default: None
+        either 'theta' or 'polarization_state' must be given.
+        polarization state with field amplitudes and phases, tuple of 4 float:
+        (E0x, E0y, Dphi, Aphi): E0X amplitde, E0Y amplitde, phase difference 
+        between X and Y components (in rad), absolute phase of plane wave (in rad).
+        The field is then calculated as E = (E0x, E0y*exp(i*Dphi*z), 0)*exp(i*Aphi*z).
+            Dphi : 
+                - positive: left hand rotating polarization
+                - negative: right hand rotating polarization 
+                - example: left circular pol. with (E0x=1, E0y=1, Dphi=np.pi/2., phi=0)
+
+    xSpot, ySpot, zSpot : float, float, float, default: 0, 0, 0
+                                       focal spot position (in nm)
+
+    kSign : int, default: -1
+               sign of wavenumber. 
+               +1: propagation from bottom to top (towards increasing z)
+               -1: propagation from top to bottom (towards smaller z, default) 
+
+      NA : float, default: 0.5
+              lens numerical aperture (NA = n sin thetamax),
+
+      f :     float, default: 100
+              lens focal distance (mm)
+
+      w0 : float, default: 1 
+              beam waist  (mm)
+
+      returnField : str, default: 'E'
+              if 'E': returns electric field; if 'B': magnetic field
+    
+    Returns
+    -------
+      E0 (B0):       Complex E-(B-)Field at each dipole position as 
+                      list of (complex) 3-tuples: [(Ex1, Ey1, Ez1), ...]    
+
+    """
+    
+    
+    if 'eps_env' in env_dict.keys():
+        cn1 = cn2 = env_dict['eps_env']**0.5
+    else:
+        cn1 = env_dict['eps1']**0.5
+        cn2 = env_dict['eps2']**0.5
+        cn3 = env_dict['eps3']**0.5
+        # spacing = env_dict['spacing']**0.5
+#        if cn1 != cn2 or cn2 != cn3:
+    if np.imag(cn1) != 0 or np.imag(cn2) != 0 or np.imag(cn3) != 0:
+        warnings.warn("Special modes only support real refractive index")
+    if cn1 != cn2 or cn2 != cn3:
+        warnings.warn("Special modes don't take reflection / transmission at interfaces into account. Results might be inaccurate.")
+    
+    cn1 = np.real(cn1)
+    cn2 = np.real(cn2)
+    cn3 = np.real(cn3)
+        
+    f *= 1E6  # conversion mm > nm
+    w0 *= 1E6  # conversion mm > nm
+        
+    xm, ym, zm = np.transpose(pos)    
+    npts = len(xm)
+    if kSign == -1 : 
+        n = cn2
+    else:    
+        n = cn1
+    k = kSign * n * (2*np.pi / wavelength)
+    prefactor = 0.5*1j*k*f**2*np.sqrt(cn1/cn2)*np.exp(-1j*k*f)/w0
+
+    Ex = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Ey = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Ez = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Bx = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    By = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    Bz = np.asfortranarray( np.zeros(len(xm)), dtype=DTYPE_C)
+    
+    for ipt in range(npts):
+        
+        x = xm[ipt] - xSpot
+        y = ym[ipt] - ySpot
+        z = zm[ipt] - zSpot
+    
+        rho, phi = _cart_cyl(x, y)
+           
+        # I_10 = _I10(x,y,z,k,n,NA,f,w0)
+        I_11 = _I11(x,y,z,k,n,NA,f,w0)
+        I_12 = _I12(x,y,z,k,n,NA,f,w0)
+     
+# Novotny (3.66)
+# Electric field
+        Ex[ipt] = prefactor * ( 1j*(I_11 + 3 * I_12)*np.sin(phi) )
+        Ey[ipt] = prefactor * (-1j*(I_11 + 3 * I_12)*np.cos(phi))
+        Ez[ipt] = 0    
+
+# Magnetic field
+
+    if returnField == 'B':
+        output = np.asfortranarray(np.transpose([Bx, By, Bz]))
+    else :
+        output = np.asfortranarray(np.transpose([Ex, Ey, Ez]))
+    ## --- return as fortran array to avoid copying of arrays in memory
+    return output 
+
+
+
+
+
+
 ##----------------------------------------------------------------------
 ## deprecated functions for backwards compatibilitiy
 ##     (may be removed in future version)
@@ -1485,7 +2438,10 @@ def evanescent_planewave(pos, env_dict, wavelength,
 
 FIELDS_LIST = [plane_wave, gaussian, 
                dipole_electric, dipole_magnetic,
-               focused_planewave, fast_electron]
+               focused_planewave, 
+               HermiteGauss00, HermiteGauss01, HermiteGauss10,
+               Radial_pol_doughnut, Azimuth_pol_doughnut,
+               fast_electron]
 
 if __name__ == "__main__":
     pass
