@@ -69,16 +69,21 @@ def _calc_repropagation(P, Escat, G_dyad_list):
 #import numba
 #@numba.njit(parallel=True)
 def _calc_extinct(wavelength, n_env, alpha_tensor, E0, E, 
-                  with_radiation_correction=False):
+                  with_radiation_correction=False, normalization_E0=False):
     """actual calculation of extinction and absorption cross sections
     
     numba compatible. tested.
     """
+    if normalization_E0:
+        I0_norm = np.sum(np.abs(E0)**2, axis=1).max()
+    else:
+        I0_norm = 1
+    
     P = np.matmul(alpha_tensor, E[...,None])[...,0]
     
-    cext = ((8 * np.pi**2 / wavelength) / n_env * 
+    cext = (1/I0_norm) * ((8 * np.pi**2 / wavelength) / n_env * 
             np.sum(np.multiply(np.conjugate(E0), P))).imag
-    cabs = ((8 * np.pi**2 / wavelength) / n_env * 
+    cabs = (1/I0_norm) * ((8 * np.pi**2 / wavelength) / n_env * 
             (np.sum(np.multiply(P, np.conjugate(E)))).imag).real
     
     if with_radiation_correction:
@@ -90,7 +95,7 @@ def _calc_extinct(wavelength, n_env, alpha_tensor, E0, E,
     return cext, csca, cabs
     
 
-def extinct(sim, field_index, with_radiation_correction=False):
+def extinct(sim, field_index, with_radiation_correction=False, normalization_E0=False):
     """Extinction, scattering and absorption cross-sections
     
     Calculates extinction, scattering and absorption crosssections
@@ -115,6 +120,9 @@ def extinct(sim, field_index, with_radiation_correction=False):
         Using the correction can lead to better agreement for very large 
         discretization stepsizes, but has only a weak influence on simulations
         with fine meshes.
+        
+    normalization_E0 : bool, default: False
+        normalizes sections to peak of incident field intensity inside structure
     
     Returns
     -------
@@ -144,7 +152,8 @@ def extinct(sim, field_index, with_radiation_correction=False):
     field_params    = sim.E[field_index][0]
     wavelength      = field_params['wavelength']
     sim.struct.setWavelength(wavelength)
-    n_env = sim.dyads.getEnvironmentIndices(wavelength, sim.struct.geometry[:1])[0]  # assume structure is fully in one environment
+    eps_env = sim.dyads.getEnvironmentIndices(wavelength, sim.struct.geometry[:1])[0]  # assume structure is fully in one environment
+    n_env = np.sqrt(eps_env)
     
     ## --- get polarizability at wavelength
     alpha_tensor = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
@@ -155,7 +164,7 @@ def extinct(sim, field_index, with_radiation_correction=False):
     E = sim.E[field_index][1]
     
     return _calc_extinct(wavelength, n_env, alpha_tensor, E0, E, 
-                         with_radiation_correction)
+                         with_radiation_correction, normalization_E0)
 
 
 
@@ -339,7 +348,7 @@ def multipole_decomp_extinct(sim, field_index, r0=None, eps_dd=0.1, quadrupoles=
     k0 = 2*np.pi / wavelength
     
     geo = sim.struct.geometry
-    n2 = sim.dyads.getEnvironmentIndices(wavelength, geo[:1])[0]  # assume structure is fully in one environment
+    n_env = sim.dyads.getEnvironmentIndices(wavelength, geo[:1])[0]  # assume structure is fully in one environment
     E = sim.E[field_index][1]
     ex, ey, ez = E.T
     
@@ -359,8 +368,8 @@ def multipole_decomp_extinct(sim, field_index, r0=None, eps_dd=0.1, quadrupoles=
     H0 = sim.efield.field_generator(r0[None,:], env_dict, returnField='H', **field_params)
     
     E2in = np.sum(np.abs(E0)**2)     # incident intensity
-    sigma_ext_p = (4 * np.pi * k0 / n2 * 1/E2in * np.sum(np.conjugate(E0)*p)).imag
-    sigma_ext_m = (4 * np.pi * k0 / n2 * 1/E2in * np.sum(np.conjugate(H0)*m)).imag
+    sigma_ext_p = (4 * np.pi * k0 / n_env * 1/E2in * np.sum(np.conjugate(E0)*p)).imag
+    sigma_ext_m = (4 * np.pi * k0 / n_env * 1/E2in * np.sum(np.conjugate(H0)*m)).imag
     
 # =============================================================================
 #     quadrupole extinction cross sections
@@ -375,7 +384,7 @@ def multipole_decomp_extinct(sim, field_index, r0=None, eps_dd=0.1, quadrupoles=
         gradE0cj = np.array([[EDXcj[1,0]-EDXcj[0,0], EDYcj[1,0]-EDYcj[0,0], EDZcj[1,0]-EDZcj[0,0]],
                              [EDXcj[1,1]-EDXcj[1,1], EDYcj[1,1]-EDYcj[0,1], EDZcj[1,1]-EDZcj[0,1]],
                              [EDXcj[1,2]-EDXcj[1,2], EDYcj[1,2]-EDYcj[0,2], EDZcj[1,2]-EDZcj[0,2]]])
-        sigma_ext_q = (np.pi * k0 / n2 * 1/(3. * E2in) * 
+        sigma_ext_q = (np.pi * k0 / n_env * 1/(3. * E2in) * 
                                np.sum(np.tensordot(gradE0cj + gradE0cj.T, q) )).imag
         
         HDXcj = np.conjugate(sim.efield.field_generator(DX_pos, env_dict, returnField='H', **field_params))
@@ -384,7 +393,7 @@ def multipole_decomp_extinct(sim, field_index, r0=None, eps_dd=0.1, quadrupoles=
         gradH0cj = np.array([[HDXcj[1,0]-HDXcj[0,0], HDYcj[1,0]-HDYcj[0,0], HDZcj[1,0]-HDZcj[0,0]],
                              [HDXcj[1,1]-HDXcj[1,1], HDYcj[1,1]-HDYcj[0,1], HDZcj[1,1]-HDZcj[0,1]],
                              [HDXcj[1,2]-HDXcj[1,2], HDYcj[1,2]-HDYcj[0,2], HDZcj[1,2]-HDZcj[0,2]]])
-        sigma_ext_mq = (2 * np.pi * k0 / n2 * 1/E2in * 
+        sigma_ext_mq = (2 * np.pi * k0 / n_env * 1/E2in * 
                                np.sum(np.tensordot(gradH0cj.T, mq) )).imag
         
         return sigma_ext_p, sigma_ext_m, sigma_ext_q, sigma_ext_mq
@@ -463,7 +472,8 @@ def farfield(sim, field_index,
                 r=100000., 
                 tetamin=0, tetamax=np.pi/2., Nteta=10, 
                 phimin=0, phimax=2*np.pi, Nphi=36, 
-                polarizerangle='none', return_value='map'):
+                polarizerangle='none', return_value='map', 
+                normalization_E0=False):
     """spatially resolved and polarization-filtered far-field scattering 
     
     For a given incident field, calculate the electro-magnetic field 
@@ -515,6 +525,9 @@ def farfield(sim, field_index,
           - "int_Es" : return the integrated scattered field (as float)
           - "int_E0" : return the integrated fundamental field (as float)
           - "int_Etot" : return the integrated total field (as float)
+          
+    normalization_E0 : bool, default: False
+        has only effect on return_value=="int_Es": Normalizes scattering to peak of incident field intensity inside structure
     
     
     Returns
@@ -607,10 +620,12 @@ def farfield(sim, field_index,
 #==============================================================================        
     ## --- fundamental field - use dummy structure with 
     env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
-    E0 = sim.efield.field_generator(_r_probe, env_dict, **field_params)
-    # ## apply correct phase for Z<0 for dipole emitter fundamental fields
-    # if sim.efield.field_generator in [fields.dipole_electric, fields.dipole_magnetic]:
-    #     E0[_r_probe.T[2]<0] *= -1
+    if return_value == 'int_Es':
+        from pyGDM2 import fields
+        E0 = fields.nullfield(_r_probe, env_dict, wavelength, returnField='E')
+    else:
+        E0 = sim.efield.field_generator(_r_probe, env_dict, **field_params)
+        
     
     ## --- electric polarization of each discretization cell via tensorial polarizability
     Eint = sim.E[field_index][1]
@@ -689,11 +704,22 @@ def farfield(sim, field_index,
     else:
         d_solid_surf = r**2 * np.sin(tetalist) * dteta * dphi
         if return_value.lower() == 'int_es':
-            return np.sum(I_sc * d_solid_surf)
+            if normalization_E0:
+                env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
+                E0 = sim.efield.field_generator(sim.struct.geometry, 
+                                                env_dict, **field_params)
+                I0_norm = np.sum(np.abs(E0)**2, axis=1).max()
+            else:
+                I0_norm = 1
+            
+            return np.sum(I_sc * d_solid_surf) / I0_norm
+        
         elif return_value.lower() == 'int_e0':
             return np.sum(I0 * d_solid_surf)
+        
         elif return_value.lower() == 'int_etot':
             return np.sum(I_tot * d_solid_surf)
+        
         else:
             raise ValueError("Parameter 'return_value' must be one of ['map', 'int_es', 'int_e0', 'int_etot'].")
 
@@ -702,15 +728,9 @@ def farfield(sim, field_index,
 
 
 
-
-
-
-
-
-
 def nearfield(sim, field_index, r_probe, which_fields=["Es","Et","Bs","Bt"],
                  val_inside_struct="field", N_neighbors_internal_field=4,
-                 method='numba'):
+                 method=''):
     """Nearfield distribution in proximity of nanostructre
     
     For a given incident field, calculate the electro-magnetic field in the 
@@ -751,9 +771,8 @@ def nearfield(sim, field_index, r_probe, which_fields=["Es","Et","Bs","Bt"],
         fields are weighted by the distance of the evaluation point to 
         the respective neighbor mesh cell.
     
-    method : str, default: "numba"
-        either "numba" or "cuda", working on CPU or CUDA-compatible GPU. Both 
-        via `numba`.
+    method : str
+        deprecated, has no effect
         
     
     Returns
@@ -779,10 +798,6 @@ def nearfield(sim, field_index, r_probe, which_fields=["Es","Et","Bs","Bt"],
 # =============================================================================
 #     Exception handling
 # =============================================================================
-    if sim.E is None: 
-        raise ValueError("Error: Scattered field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
-    
-    
     if str(val_inside_struct).lower() == 'none':
         try:
             import scipy
@@ -800,6 +815,9 @@ def nearfield(sim, field_index, r_probe, which_fields=["Es","Et","Bs","Bt"],
 # =============================================================================
 #     preparation
 # =============================================================================
+    from pyGDM2 import tools
+
+
     if len(np.shape(r_probe)) == 1:
         if len(r_probe) == 3:
             r_probe = [[r_probe[0]], [r_probe[1]], [r_probe[2]]]
@@ -815,105 +833,127 @@ def nearfield(sim, field_index, r_probe, which_fields=["Es","Et","Bs","Bt"],
                          "elements, either floats, or lists.")
     r_probe = np.transpose(r_probe)
     
-    field_kwargs    = sim.E[field_index][0]
+    field_kwargs    = tools.get_field_indices(sim)[field_index]
     wavelength      = field_kwargs['wavelength']
     
     sim.struct.setWavelength(wavelength)
     conf_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
     
     
-# =============================================================================
-#     evaluate Green's function and propagate fields
-# =============================================================================
-    ## --- fundamental field - use dummy structure with 
+    ## --- fundamental field evaluation - use dummy structure with 
     env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
     E0 = sim.efield.field_generator(r_probe, env_dict, **field_kwargs)
-    
-    env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
     B0 = sim.efield.field_generator(r_probe, env_dict, returnField='H', **field_kwargs)
     
+    
     ## --- electric polarization of each discretization cell via tensorial polarizability
     if "es" in which_fields or "et" in which_fields or \
                 "bs" in which_fields or "bt" in which_fields or \
                 "hs" in which_fields or "ht" in which_fields:
+        if sim.E is None: 
+            raise ValueError("Error: Scattered field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
         Eint = sim.E[field_index][1]
         alpha_tensor = sim.dyads.getPolarizabilityTensor(wavelength, sim.struct)
         P = np.matmul(alpha_tensor, Eint[...,None])[...,0]
     
-    
-    ## +++++++++++++ electric field +++++++++++++
-    if "es" in which_fields or "et" in which_fields:
-        if method.lower()=='numba':
-            ## Greens function for each dipole
-            G_NF = np.zeros((len(sim.struct.geometry), len(r_probe), 3, 3), 
-                            dtype = sim.efield.dtypec)
-            sim.dyads.eval_G(sim.struct.geometry, r_probe, 
-                             sim.dyads.G_EE, wavelength, conf_dict, G_NF)
-            
-            ## propagate fields 
-            G_NF = np.moveaxis(G_NF, 0,2).reshape((len(r_probe), 3, -1))  # re-arrange for direct matrix-vector multiplication
+        
+        ## --- init empty numpy arrays
+        if "es" in which_fields or "et" in which_fields:
             Escat = np.zeros(shape=(len(r_probe), 3), dtype=sim.efield.dtypec)
-            _calc_repropagation(P, Escat, G_NF)
-            
-    ## +++++++++++++ magnetic field +++++++++++++
-    if "bs" in which_fields or "bt" in which_fields or "hs" in which_fields or "ht" in which_fields:
-        if method.lower()=='numba':
-            ## Greens function for each dipole
-            G_NF = np.zeros((len(sim.struct.geometry), len(r_probe), 3, 3), 
-                            dtype = sim.efield.dtypec)
-            sim.dyads.eval_G(sim.struct.geometry, r_probe, 
-                             sim.dyads.G_HE, wavelength, conf_dict, G_NF)
-            
-            ## propagate fields 
-            G_NF = np.moveaxis(G_NF, 0,2).reshape((len(r_probe), 3, -1))  # re-arrange for direct matrix-vector multiplication
+        if "bs" in which_fields or "bt" in which_fields or "hs" in which_fields or "ht" in which_fields:
             Bscat = np.zeros(shape=(len(r_probe), 3), dtype=sim.efield.dtypec)
-            _calc_repropagation(P, Bscat, G_NF)
-            
+        
 # =============================================================================
 #     treat positions inside structure
 # =============================================================================
-    val_inside_struct = str(val_inside_struct)
-    if val_inside_struct.lower() != "none":
-#    if str(val_inside_struct).lower() != "field":
-        from scipy.linalg import norm
-        for i, R in enumerate(r_probe):
-            dist_list = norm(sim.struct.geometry - R, axis=1)
-            idcs_min_dist = np.argsort(dist_list)
-            ## --- if inside, replace fields
-            if abs(dist_list[idcs_min_dist[0]]) <= 1.005*sim.struct.step:
-                if val_inside_struct.lower() == "nan":
-                    fill_valueE = np.nan
-                    fill_valueB = np.nan
-                elif val_inside_struct.lower() == "field":  # calc internal scattered field
-                    fill_valueE = np.average(Eint[idcs_min_dist[:N_neighbors_internal_field]], 
-                                 weights = 1/(sim.struct.step/100. + dist_list[idcs_min_dist[:N_neighbors_internal_field]]**1), 
-                                 axis=0) - E0[i]  # subtract E0 --> scattered field only
-                    if sim.H is not None:
-                        Hint = sim.H[field_index][1]
-                        fill_valueB = np.average(Hint[idcs_min_dist[:N_neighbors_internal_field]], 
-                                 weights = 1/(sim.struct.step/100. + dist_list[idcs_min_dist[:N_neighbors_internal_field]]**1), 
-                                 axis=0) - B0[i]  # subtract B0 --> scattered field only
-                    else:
-                        warnings.warn("No H-values inside structure calculated. Setting internal magnetic field zero. " + 
-                                      "Please run 'scatter' with the according parameter.")
-                        fill_valueB = 0   # here we need the B-field inside. Not yet implemented
-                else:
-                    fill_valueE = 0
-                    fill_valueB = 0
-                if "es" in which_fields or "et" in which_fields:
-                    Escat[i] = fill_valueE
-                if "bs" in which_fields or "bt" in which_fields or "hs" in which_fields or "ht" in which_fields:
-                    Bscat[i] = fill_valueB
-    
-    if "et" in which_fields:
-        Etot = Escat + E0
-    if "bt" in which_fields or "ht" in which_fields:
-        Btot = Bscat + B0
-    
+        val_inside_struct = str(val_inside_struct)
         
+        _i_r_probe_out = []
+        _r_probe_out = []
+        if val_inside_struct.lower() != "none":
+            from scipy.linalg import norm
+            for i, R in enumerate(r_probe):
+                dist_list = norm(sim.struct.geometry - R, axis=1)
+                idcs_min_dist = np.argsort(dist_list)
+                ## --- if inside, replace fields
+                if abs(dist_list[idcs_min_dist[0]]) <= 1.005*sim.struct.step:
+                    if val_inside_struct.lower() == "nan":
+                        fill_valueE = np.nan
+                        fill_valueB = np.nan
+                    elif val_inside_struct.lower() == "field":  # calc internal scattered field
+                        fill_valueE = np.average(Eint[idcs_min_dist[:N_neighbors_internal_field]], 
+                                     weights = 1/(sim.struct.step/100. + dist_list[idcs_min_dist[:N_neighbors_internal_field]]**1), 
+                                     axis=0) - E0[i]  # subtract E0 --> scattered field only
+                        if sim.H is not None:
+                            Hint = sim.H[field_index][1]
+                            fill_valueB = np.average(Hint[idcs_min_dist[:N_neighbors_internal_field]], 
+                                     weights = 1/(sim.struct.step/100. + dist_list[idcs_min_dist[:N_neighbors_internal_field]]**1), 
+                                     axis=0) - B0[i]  # subtract B0 --> scattered field only
+                        else:
+                            if "bs" in which_fields or "bt" in which_fields or "hs" in which_fields or "ht" in which_fields:
+                                warnings.warn("No H-values inside structure calculated. Setting internal magnetic field zero. " + 
+                                              "Please run 'scatter' with the according parameter.")
+                            fill_valueB = 0   # here we need the B-field inside. Not yet implemented
+                    else:
+                        fill_valueE = 0
+                        fill_valueB = 0
+                    if "es" in which_fields or "et" in which_fields:
+                        Escat[i] = fill_valueE
+                    if "bs" in which_fields or "bt" in which_fields or "hs" in which_fields or "ht" in which_fields:
+                        Bscat[i] = fill_valueB
+                ## --- not inside, requires field propagation
+                else:
+                    _i_r_probe_out.append(i)
+                    _r_probe_out.append(R)
+        
+        _r_probe_out = np.array(_r_probe_out)
+        _i_r_probe_out = np.array(_i_r_probe_out)
+            
+# =============================================================================
+#     evaluate Green's function and propagate fields
+# =============================================================================
+        if len(_r_probe_out) > 0:
+            ## +++++++++++++ electric field +++++++++++++
+            if "es" in which_fields or "et" in which_fields:
+                ## Greens function for each dipole
+                G_NF = np.zeros((len(sim.struct.geometry), len(_r_probe_out), 3, 3), 
+                                dtype = sim.efield.dtypec)
+                sim.dyads.eval_G(sim.struct.geometry, _r_probe_out, 
+                                 sim.dyads.G_EE, wavelength, conf_dict, G_NF)
+                
+                ## propagate fields 
+                G_NF = np.moveaxis(G_NF, 0,2).reshape((len(_r_probe_out), 3, -1))  # re-arrange for direct matrix-vector multiplication
+                _Escat_out = np.zeros(shape=(len(_r_probe_out), 3), dtype=sim.efield.dtypec)
+                _calc_repropagation(P, _Escat_out, G_NF)
+                
+                ##  fill final field-list with outside values
+                Escat[_i_r_probe_out] = _Escat_out
+                
+            ## +++++++++++++ magnetic field +++++++++++++
+            if "bs" in which_fields or "bt" in which_fields or "hs" in which_fields or "ht" in which_fields:
+                ## Greens function for each dipole
+                G_NF = np.zeros((len(sim.struct.geometry), len(_r_probe_out), 3, 3), 
+                                dtype = sim.efield.dtypec)
+                sim.dyads.eval_G(sim.struct.geometry, _r_probe_out, 
+                                 sim.dyads.G_HE, wavelength, conf_dict, G_NF)
+                
+                ## propagate fields 
+                G_NF = np.moveaxis(G_NF, 0,2).reshape((len(_r_probe_out), 3, -1))  # re-arrange for direct matrix-vector multiplication
+                _Bscat_out = np.zeros(shape=(len(_r_probe_out), 3), dtype=sim.efield.dtypec)
+                _calc_repropagation(P, _Bscat_out, G_NF)
+                
+                ##  fill final field-list with outside values
+                Bscat[_i_r_probe_out] = _Bscat_out
+    
 # =============================================================================
 #     bundle output
 # =============================================================================
+        if "et" in which_fields:
+            Etot = Escat + E0
+        if "bt" in which_fields or "ht" in which_fields:
+            Btot = Bscat + B0
+    
+    
     return_field_list = []
     for f_type in which_fields:
         
@@ -1042,10 +1082,10 @@ def field_gradient(sim, field_index,
     r_probe_px.T[0] += delta
     r_probe_mx.T[0] -= delta
     F_px = nearfield(sim, field_index, r_probe_px, which_fields)
-    F_Mx = nearfield(sim, field_index, r_probe_mx, which_fields)
+    F_mx = nearfield(sim, field_index, r_probe_mx, which_fields)
 
     dFdx = []
-    for Epx, Emx in zip(F_px, F_Mx):
+    for Epx, Emx in zip(F_px, F_mx):
         dFdx.append(np.array([np.ravel((Epx.T[3:].T - Emx.T[3:].T) / (2.*delta))]))
 
     ## --- d/dy
@@ -1054,10 +1094,10 @@ def field_gradient(sim, field_index,
     r_probe_py.T[1] += delta
     r_probe_my.T[1] -= delta
     F_py = nearfield(sim, field_index, r_probe_py, which_fields)
-    F_My = nearfield(sim, field_index, r_probe_my, which_fields)
+    F_my = nearfield(sim, field_index, r_probe_my, which_fields)
     
     dFdy = []
-    for Epy, Emy in zip(F_py, F_My):
+    for Epy, Emy in zip(F_py, F_my):
         dFdy.append(np.array([np.ravel((Epy.T[3:].T - Emy.T[3:].T) / (2.*delta))]))
     
     ## --- d/dz    
@@ -1066,11 +1106,11 @@ def field_gradient(sim, field_index,
     r_probe_pz.T[2] += delta
     r_probe_mz.T[2] -= delta
     F_pz = nearfield(sim, field_index, r_probe_pz, which_fields)
-    F_Mz = nearfield(sim, field_index, r_probe_mz, which_fields)
+    F_mz = nearfield(sim, field_index, r_probe_mz, which_fields)
     
     ## --- assemble
     dFdz = []
-    for Epz, Emz in zip(F_pz, F_Mz):
+    for Epz, Emz in zip(F_pz, F_mz):
         dFdz.append(np.array([np.ravel((Epz.T[3:].T - Emz.T[3:].T) / (2.*delta))]))
         
         
@@ -1108,21 +1148,21 @@ def optical_force(sim, field_index,
     return_value : str, default: 'total'
         Values to be returned. Either 'total' (default) or 'structure'.
         
-        - "total" : return the total deposited heat in nW (float)
+        - "total" : return the total force on the entire particle (3-vector [Fx, Fy, Fz])
         
         - "structure" : return spatially resolved force, in other words 
-        the force acting on each meshpoint. --> list of tuples [x,y,z,q])
+        the force acting on each meshpoint. --> list of tuples [x,y,z,F_ix,F_iy,F_iz])
     
     
     Returns
     -------
     F : tuple (Fx,Fy,Fz) *or* list of force vectors [x,y,z,Fx,Fy,Fz]
     
-        - `return_value`=="total" : (return float)
+        - `return_value`=="total" : (3-vector [Fx, Fy, Fz])
           total force on structure 
 
-        - `return_value`=="structure" : (return list of tuples [x,y,z,q])
-          force at each mesh-cell positions [x,y,z].
+        - `return_value`=="structure" : (return list of tuples [xi,yi,zi,F_ix,F_iy,F_iz])
+          force at all mesh-cells `i` of positions [xi,yi,zi].
     
     
     Notes
