@@ -4,8 +4,8 @@
 #
 #This program is free software: you can redistribute it and/or modify
 #it under the terms of the GNU General Public License as published by
-#the Free Software Foundation, either version 3 of the License, or
-#(at your option) any later version.
+#the Free Software Foundaeon, either version 3 of the License, or
+#(at your opeon) any later version.
 #
 #This program is distributed in the hope that it will be useful,
 #but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -33,7 +33,7 @@ class _interp1dPicklable:
     **Note:** there might be still pickle-problems with certain c / fortran 
     wrapped libraries
     
-    From: http://stackoverflow.com/questions/32883491/pickling-scipy-interp1d-spline
+    From: http://stackoverflow.com/quesoons/32883491/pickling-scipy-interp1d-spline
     """
     def __init__(self, xi, yi, **kwargs):
         from scipy.interpolate import interp1d
@@ -565,9 +565,27 @@ class tio2(object):
     
     model for visible and NIR range (~500nm - ~1500nm)
     from https://refractiveindex.info/?shelf=main&book=TiO2&page=Devore-o
+    
+    Parameters
+    ----------
+        orientation : str, default: 'avg'
+          - 'avg' (or 'mean'): average of n_o and n_e: n = (2*n_o + n_e) / 3
+          - 'n_o' : ordinary axis permittivity
+          - 'n_e' : extra-ordinary axis permittivity
+    
     """
-    def __init__(self):
+    def __init__(self, orientation='avg'):
         self.__name__ = 'TiO2 (Devore)'
+        self.orientation = orientation.lower()
+    
+    
+    def _n_o(self, wavelength):
+        n = np.sqrt(5.913 + 0.2441 / ((wavelength/1E3)**2 - 0.0803))
+        return n
+    
+    def _n_e(self, wavelength):
+        n = np.sqrt(7.197 + 0.3322 / ((wavelength/1E3)**2 - 0.0843))
+        return n
     
     def epsilon(self, wavelength):
         """TiO2 dielectric function
@@ -578,8 +596,15 @@ class tio2(object):
             wavelength (in nm) at which to evaluate dielectric function
         
         """
-        eps = 5.913 + 0.2441 / ((wavelength/1E3)**2 - 0.0803)
-        eps = complex(eps)
+        if self.orientation in ['n_o', 'no']:
+            eps = self._n_o(wavelength)**2  + 0j  # purely real in available spectral range
+        elif self.orientation in ['n_e', 'ne']:
+            eps = self._n_e(wavelength)**2 + 0j
+        else:
+            n_o = self._n_o(wavelength)
+            n_e = self._n_e(wavelength)
+            n_avg = (2*n_o + n_e) / 3.
+            eps = n_avg**2 + 0j
         
         return eps
 
