@@ -1733,7 +1733,8 @@ def _meshCubic(XRANGE, YRANGE, ZRANGE, condition, Z_offset=0.5):
     return sp
 
 
-def _meshHexagonalCompact(XRANGE, YRANGE, ZRANGE, condition, ORIENTATION=1, Z_offset=0.5):
+def _meshHexagonalCompact(XRANGE, YRANGE, ZRANGE, condition, ORIENTATION=1, Z_offset=0.5,
+                          auto_expand_range=True):
     """mesh on a hexagonal compact grid
     
     Parameters
@@ -1752,6 +1753,9 @@ def _meshHexagonalCompact(XRANGE, YRANGE, ZRANGE, condition, ORIENTATION=1, Z_of
     
     Z_offset : float, default: 0.5
         additional z-offset, to avoid placing dipoles at z=0
+    
+    auto_expand_range : bool, default: True
+        automatically expand meshing range in order to account for denser than cubic meshing
         
     Returns
     -------
@@ -1759,18 +1763,20 @@ def _meshHexagonalCompact(XRANGE, YRANGE, ZRANGE, condition, ORIENTATION=1, Z_of
          list of 3d coordinate-index tuples
     """
     sp = []
-    ## Correct range for inhomogeneous hex-meshcell extensions
     XRANGE = list(XRANGE)
     YRANGE = list(YRANGE)
     ZRANGE = list(ZRANGE)
-    XRANGE[0] = XRANGE[0] - np.abs(XRANGE[0])*2/np.sqrt(3)
-    XRANGE[1] = XRANGE[1] + np.abs(XRANGE[1])*2/np.sqrt(3)
     
-    YRANGE[0] = YRANGE[0] - np.abs(YRANGE[0])*2/np.sqrt(3)
-    YRANGE[1] = YRANGE[1] + np.abs(YRANGE[1])*2/np.sqrt(3)
-    
-    ZRANGE[0] = ZRANGE[0] - np.abs(ZRANGE[0])*2/np.sqrt(3)
-    ZRANGE[1] = ZRANGE[1] + np.abs(ZRANGE[1])*2/np.sqrt(3)
+    ## Auto-expand range for inhomogeneous hex-meshcell extensions
+    if auto_expand_range:
+        XRANGE[0] = XRANGE[0] - np.abs(XRANGE[0])*2/np.sqrt(3)
+        XRANGE[1] = XRANGE[1] + np.abs(XRANGE[1])*2/np.sqrt(3)
+        
+        YRANGE[0] = YRANGE[0] - np.abs(YRANGE[0])*2/np.sqrt(3)
+        YRANGE[1] = YRANGE[1] + np.abs(YRANGE[1])*2/np.sqrt(3)
+        
+        ZRANGE[0] = ZRANGE[0] - np.abs(ZRANGE[0])*2/np.sqrt(3)
+        ZRANGE[1] = ZRANGE[1] + np.abs(ZRANGE[1])*2/np.sqrt(3)
     
     
     if ORIENTATION==1:
@@ -1887,6 +1893,7 @@ def image_to_struct(img_name, nm_per_pixel, stepsize, H, threshold=100,
         img.load()   # required for png.split()
     else:
         img = Image.fromarray(img_name)
+        
 
     ## --- remove alpha channel, if exists
     if len(img.split())==4:
@@ -1897,6 +1904,20 @@ def image_to_struct(img_name, nm_per_pixel, stepsize, H, threshold=100,
         ## RGB data --> no alpha channel
         background = img
     
+    
+    ## --- hex meshing: add white border to account for denser mesh than cubic
+    if mesh.lower() == 'hex':
+        old_size = img.size
+        new_size = (int(old_size[0]*1.41), int(old_size[1]*1.41))
+        new_im = Image.new("RGB", new_size, color=(255, 255, 255))
+        # new_im = Image.new("RGB", new_size, color=0)
+        new_im.paste(background, (int((new_size[0]-old_size[0])*0), 
+                                  int((new_size[1]-old_size[1])/1) ))
+        # new_im.paste(background, ((new_size[0]-old_size[0])//2,
+        #                           (new_size[1]-old_size[1])//2))
+        background = new_im
+    
+    
     ## original image to return if required
     img0 = background.convert('L')
     data0 = np.asarray(img0.getdata()).reshape(img0.size[::-1])
@@ -1929,7 +1950,7 @@ def image_to_struct(img_name, nm_per_pixel, stepsize, H, threshold=100,
     if mesh == 'cube':
         sp = _meshCubic([0, data.shape[0]], [0, data.shape[1]], [0, H], condition)
     elif mesh == 'hex':
-        sp = _meshHexagonalCompact([0, data.shape[0]], [0, data.shape[1]], [0, H], condition, ORIENTATION)
+        sp = _meshHexagonalCompact([0, data.shape[0]], [0, data.shape[1]], [0, H], condition, ORIENTATION, auto_expand_range=False)
     sp = np.array(sp, dtype=np.float).T*stepsize
     sp[2] -= np.min(sp[2])  # shift Z so that bottom at z=0
     sp[2] += stepsize/2         # shift Z so that bottom at z=step/2
