@@ -897,6 +897,7 @@ def get_mesh_from_geometry(geo, N_try=20, verbose=0):
         11.0:"hex",   # 2-layer 3D
         11.20:"hex",   # 2D - config1
         11.28:"hex",   # 2D - config2
+        13.94:"hex",   # 2D - config3
         }
     
     if len(geo) <= 1:
@@ -919,7 +920,7 @@ def get_mesh_from_geometry(geo, N_try=20, verbose=0):
         if verbose:
             print("step= {:.1f}, distsum={:.2f} --> mesh: '{}'".format(step, distsum, meshtype))
     except KeyError:
-        warnings.warn("Mesh not detected, falling back to 'cubic'.")
+        warnings.warn("Mesh not detected, falling back to 'cubic'. mindist value = {}".format(distsum))
         meshtype = 'cube'
     
     return meshtype
